/* Tibi - Chinese handwriting recognition
 * Copyright (C) 2007 Stefan Dirnstorfer <stefan@dadim.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
package matcher;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Vector;

import nodes.NodeList;
import nodes.NodeListTraverser;
import rules.PatternTraverser;
import rules.Rule;
import util.BiStack;

public final class ContextController {
	
	static double base= 0.95;
	
	BiStack<ScannerContext> active[];
	int activeBucket, activeCount;
	Map<String, ScannerContext> compmap;
	Vector<ScannerContext> waiting;
	double bestFits[];
	String bestChars[];
		
	public ContextController(int size) {
		bestFits= new double[size];
		bestChars= new String[size];
	}
	
	public void clear() {
		active= new BiStack[(int) (Math.log(ParameterSet.minimumFit)/Math.log(base))];
		activeBucket=0;
		activeCount=0;
		for (int i=0; i<bestFits.length; ++i) {
			bestFits[i]=0.0;
			bestChars[i]=null;
		}
		waiting= new Vector<ScannerContext>();
		compmap= new HashMap<String, ScannerContext>();
	}

	public void addContext(ScannerContext context) {		
		for (ScannerContext child : context.getOffspring()) {
			addContext(child);
		}
		if (context.getFit() > ParameterSet.minimumFit) {
			if (context.isActive()) {
				addActiveContext(context);
			} else {
				if (context.isComplete())
					addCompletedContext(context);
				else
					addPendingContext(context);
			}
		}
	}

	public void addActiveContext(ScannerContext context) {
		int bucket= (int) (Math.log(context.getFit())/Math.log(base));
		if (bucket<0)
			bucket=0;
		if (bucket<active.length) {
			 if (active[bucket]==null)
				active[bucket]= new BiStack<ScannerContext>();
			active[bucket].addFirst(context);
			if (bucket<activeBucket)
				activeBucket=bucket;
			activeCount++;
		}
	}

	public void addContext(Rule rule, NodeList nodes) {
		addContext(new ScannerContext(rule.getTarget(), 
				   new PatternTraverser(rule.getPattern()), 
				   new NodeListTraverser(nodes)));
	}
	
	public void addPendingContext(ScannerContext context) {
		waiting.add(context);
	}
	
	public void addCompletedContext(ScannerContext context) {
		ScannerContext present= compmap.get(context.target);
		if (present != null) {
			if (present.getFit()>context.getFit()) {
				//return;
			}
			//context.target="."+context.target;
		}
		//context.target=context.target+(context.number % 10);
		compmap.put(context.target, context);
		
		if (context.fit > bestFits[0]) {
			double h= context.fit;
			String k= context.target;
			int i=bestFits.length-1;
			while (i>=0 && bestFits[i]>=h && k!=bestChars[i]) { 
				i--;
			}
			while (i>=0) {
				double h2= bestFits[i];
				String k2= bestChars[i];
				bestFits[i]=h;
				bestChars[i]=k;
				if (k2==context.target) {
					break;
				}
				h=h2;
				k=k2;
				i--;
			}
		}
	}
	
	public ScannerContext nextActiveContext() {
		while (activeBucket < active.length) {
			if (active[activeBucket]!=null)
				if (active[activeBucket].hasNext()) {
					activeCount--;
					ScannerContext context= active[activeBucket].poll();
					return context;
				}
			activeBucket++;
		}
		return null;
	}
	
	public void reActivate(double completedCost) {
		Collection<ScannerContext> oldwaiting=waiting;
		waiting=new Vector<ScannerContext>();
		for (int i=0; i<bestFits.length; ++i) {
			bestFits[i]*= completedCost;
		}
		for (ScannerContext complete : compmap.values()) {
			complete.applyFit(completedCost);
		}
		for (ScannerContext context : oldwaiting) {
			addContext(context);
		}
	}
	
	public double bestFit() {
		double bestfit= bestFits[bestFits.length-1];
		if (bestfit==0.0)
			return 1.0;
		return bestfit;
	}
	
	public int matchCount() {
		double ref= Math.pow(base, activeBucket);
		//System.out.println("ref= "+ref);
		int i=bestFits.length-1;
		int n=0;
		while (i>=0 && bestFits[i]>ref && bestChars[i]!=null) {
			//System.out.println("found "+bestFits[i]);
			i--;
			n++;
		}
		return n;
	}
	
	public String getMatch(int i) {
		if (getFit(i)< ParameterSet.worstSecondBest * getFit(0))
			return null;
		return bestChars[bestChars.length-1-i];
	}
	
	public double getFit(int i) {
		return bestFits[bestChars.length-1-i];
	}
	
	public boolean isActive() {
		if (activeCount==0)
			return false;
		if (bestChars[0]==null)
			return true;
		double ref= Math.pow(base, activeBucket+1)+1e-8;
		if (ref > ParameterSet.worstSecondBest*bestFits[bestFits.length-1]
		    && 
		    ref > bestFits[0])
			return true;
		return false;
	}

}
