/* Tibi - Chinese handwriting recognition
 * Copyright (C) 2007 Stefan Dirnstorfer <stefan@dadim.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
package matcher;

import java.util.Collection;
import java.util.LinkedList;
import java.util.Vector;

import nodes.Node;
import nodes.NodeList;
import nodes.NodeTraverser;
import nodes.Vector2d;
import rules.Rule;
import rules.RuleSet;
import rules.StrokePattern;

public class HanziScanner implements Runnable {

	RuleSet rules;
	LinkedList<TibiListener> listener;
	ContextController candidates;
	NodeList nodeList;
	int notified;
	boolean stopped, running;
	
	public HanziScanner(RuleSet rules, int n) {
		this.rules= rules;
		listener=new LinkedList<TibiListener>();
		candidates= new ContextController(n);
		clear();		
	}
	
	public void clear() {
		candidates.clear();
		nodeList= new NodeList();
		for (Rule r : rules) {
			if (r.isVisible()) {
				candidates.addContext(r, nodeList);
			}
		}		
		for (TibiListener l : listener) {
			l.status(TibiListener.CLEAR, null);
		}
	}
	
	public void stopNextOccasion() {
		stopped= true;
		if (running)
			for (TibiListener l : listener)
				l.status(TibiListener.STOP,null);
	}
	
	public boolean isStopped() {
		return stopped;
	}
	
	public void addListener(TibiListener listener) {
		this.listener.add(listener);
	}
	
	public void printContext(String literal) {
		System.out.println(candidates.compmap.get(literal));
	}
	
	public void addStrokeList(Collection<Vector<Vector2d>> path) {
	   	for (Vector<Vector2d> v: path) {
	   		addStroke(v);
	   	}
	}
	   	 
	public void addStroke(Vector<Vector2d> path) {
		Vector2d vec = new Vector2d();
		double minskiplen= ParameterSet.minimalFirstStrokeLength;
		
		int last=0;
    	int lastdir=0;
    	nodeList.startStroke();
    	
   		for (int i=0; i<path.size(); ++i) {
   			vec.set(path.elementAt(i));
			vec.sub(path.elementAt(last));
		 	if (vec.length() > minskiplen) {
		 		minskiplen = ParameterSet.minimalStrokeLength;
		 		int curdir=Node.getStrokeType(vec, lastdir);
		 		if (curdir!=lastdir) {
		 			lastdir=curdir;
		 			nodeList.add(path.elementAt(last), curdir);
		 		}
		 		while (vec.length() > minskiplen) {
		 			last++;
		   			vec.set(path.elementAt(last));
					vec.sub(path.elementAt(i));		 			
		 		}
		 	}
		} 
		nodeList.endStroke(path.lastElement());
		candidates.reActivate(ParameterSet.unmatchedStroke);
	}
	
	public synchronized void run() {
		running=true;
		stopped= false;
		notified=0;
		for (TibiListener l : listener)
			l.status(TibiListener.START, null);
		if (ParameterSet.verbose)
			System.out.println("--- run ---\n"+nodeList);
		try {
			while (candidates.isActive() && !isStopped()) {
				ScannerContext context= candidates.nextActiveContext();
					
				match(context);

				candidates.addContext(context);

				int n= candidates.matchCount();
				while (n>notified) {
					if (ParameterSet.verbose) System.out.println("Notify "+notified+": "+rules.getLineNumber(candidates.getMatch(notified))+" with "+candidates.getFit(notified));
					for (TibiListener l : listener)
						l.setResult(notified, candidates.getMatch(notified));
					notified++;
				}	
			}
			running=false;
			// if ended normally, notify and print remaining matches
			if (!stopped) {
				while (notified< candidates.bestFits.length) {
					for (TibiListener l : listener)
						l.setResult(notified, candidates.getMatch(notified));
					notified++;
				}
				for (TibiListener l : listener)
					l.status(TibiListener.FINISHED, null);
			}
		} catch (Error e) {
			for (TibiListener l : listener)
				l.status(TibiListener.ERROR, e.getMessage());
		}
	}

	void match(ScannerContext context) {
		ScannerContext variant=null;
		if (context.nodes.hasVariations()) {
			context.nodes.variationsChecked();
			variant= context.spawn();
		}
		NodeTraverser nodes= context.nodes.next();
		context.message("--- new stroke:"+context.nodes+" ---");
		match(context, nodes, true, false);

		if (variant!=null) {
			if (ParameterSet.reverseStroke && context.getFit()<ParameterSet.minimumFit) {
				// Testing variant with reversed orientation.
				ScannerContext variant2= variant.spawn();
				variant2.nodes.invert();
				variant2.message("--- Reversed Stroke direction ---");
				variant2.applyFit(ParameterSet.reversedStroke);
			}
			if (ParameterSet.permuteStrokeOrder && !context.nodes.isInverted()) {
				// Testing variant with incorrect stroke order
				variant.message("--- skipped Stroke ---");
				variant.message("Old stroke: ",variant.nodes.toString());
				variant.nodes.skip();
				variant.message("Next stroke: ",variant.nodes.toString());
				variant.applyFit(ParameterSet.skippedStroke);
			} else 
				variant.applyFit(0);
		}
	}
	
	void match(ScannerContext context, NodeTraverser nodes, boolean forcePrefix, boolean matched) {

		context.message("Expected stroke: "+StrokePattern.toString(context.searchedStrokeTypes()));

		Collection<ScannerContext> parasites;
		Vector2d vec= new Vector2d();
		Vector2d lastmatch= null;
		double waste= 0;

		if (context.strokeIsConnected() && !matched) {
			vec.set(context.getLastPoint());
			vec.sub(nodes.first());
			context.message("Broken line: ",vec.length());
			context.applyFit(Math.pow(ParameterSet.lineBreak,vec.length()));
		}
			
		while (nodes.next() && context.getFit() > ParameterSet.minimumFit) {
			
			// unmatched length too big, then cancel
			if (!matched && waste>ParameterSet.maximumInitialJunk)
				break;
			
			// don't match multiple lines in inverted mode
			if (nodes.isInverted() && matched && !context.strokeIsConnected()) {
				while (nodes.next()) {
					vec.set(nodes.strokeStart());
					vec.sub(nodes.strokeEnd());
					waste+=vec.length();
				};
				break;
			}

			Vector2d node1= nodes.strokeStart();
		
			if (nodes.matchesStrokeType(context.searchedStrokeTypes())) {
				// finish the stroke, without trying to match anything else.
				if (matched && !context.strokeIsConnected()) {
					double vwaste=waste;
					ScannerContext variant= context.spawn();
					NodeTraverser vnodes= new NodeTraverser(nodes);
					while (vnodes.next() && vnodes.isMatched()) {};
					while (vnodes.next()) {
						vec.set(vnodes.strokeStart());
						vec.sub(vnodes.strokeEnd());
						vwaste+=vec.length();
					};
					variant.message("Skipping matched stroke part ("+vwaste+")");
					variant.applyFit(Math.pow(ParameterSet.redundantStroke, vwaste));
					//match(variant,nodes.copy(), forcePrefix,matched);	
				}
				Vector2d node2= nodes.strokeEnd(context.searchedStrokeTypes());
		
				// compute difference to straight connection and punish
				if (waste>0) {
					if (lastmatch==null) {
						vec.set(0,0);
					} else {
						vec.set(lastmatch);
						vec.sub(node1);
					}
					double fit= Math.pow(ParameterSet.redundantStroke,Math.max(0.0,waste-vec.length()));
					context.message("Non straight connection: ",fit," ("+vec.length()+")");
					context.applyFit(fit);
				}
				
				context.message("Matched: ",StrokePattern.toString(context.searchedStrokeTypes())," at ",node1," - ",node2);
	
				context.expand(node1);
				context.expand(node2);
				matched=true;
				lastmatch=node2;
				waste=0;

				parasites= context.nextStrokePattern(true);
				// treat all triggered derived characters
				for (ScannerContext variant : parasites) {
					match(variant, nodes.copy(), false, true);
				}
			} else if (!nodes.isMatched()) {
				Vector2d node2= nodes.strokeEnd();
				vec.set(node1);
				vec.sub(node2);
				context.message("Unmatched: ",StrokePattern.toString(context.searchedStrokeTypes())," at ",node1," - ",node2," (",vec.length(),")");
				waste= waste+ vec.length();
			}
		}
		
		//if nothing was matched at all
		if (!matched) {
			context.message("Nothing matched");
			if (!forcePrefix || !context.hasMorePattern())
				context.applyFit(ParameterSet.unmatchedStroke);
			else {
				// Skip the expected stroke and search for the next on
				context.applyFit(ParameterSet.missingPrefix);
				context.expand(nodes.first());
				context.nextStrokePattern(false);
				nodes.clear();
				// if stroke has a connction, allow prefix to be missing, die otherwise
				if (context.strokeIsConnected()) {
					context.message("Test: Skipping prefix ");
					match(context,nodes,false, false);
				} else {
					context.message("Unmatched stroke.");
					context.applyFit(ParameterSet.unmatchedStroke);
				}
			}
		} else {
			if (waste>0) context.message("Redundant suffix found");
			context.applyFit(Math.pow(ParameterSet.redundantStroke, waste));
			// a connecting stroke is expected, but missing.
			missingSuffix(context, nodes.last());
		}
	}
	
	void missingSuffix(ScannerContext context, Vector2d last) {
		if (context.strokeIsConnected()) {
			ScannerContext variant= context.spawn();
			while (variant.hasMorePattern() && variant.strokeIsConnected()) {
				variant.expand(last);
				for (ScannerContext var2 : variant.nextStrokePattern(true))
					missingSuffix(var2, last);
				variant.message("Missing Suffix assumed");
				variant.applyFit(ParameterSet.missingSuffix);
			}
		}		
	}
}
