/* Tibi - Chinese handwriting recognition
 * Copyright (C) 2007 Stefan Dirnstorfer <stefan@dadim.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
package matcher;

import java.util.Collection;
import java.util.LinkedList;

import nodes.NodeListTraverser;
import nodes.Vector2d;

import rules.ArgumentPattern;
import rules.Locator;
import rules.MacroPattern;
import rules.Parasite;
import rules.PatternTraverser;
import rules.ReferencePattern;
import rules.Rule;
import rules.StrokePattern;
import util.BiStack;
import util.MessageStack;

public final class ScannerContext implements Comparable<ScannerContext>{

	static long numberCounter=0;
	
	long number;
	double fit;
	BiStack<BiStack<ScaledLocator>> locators;
	BiStack<BoundingBox> bounds;
	Collection<ScannerContext> offspring;
	MessageStack messages;
	
	PatternTraverser pattern;
	NodeListTraverser nodes;
	Vector2d lastpoint;
	String target;
	
	public ScannerContext() {
		offspring= new LinkedList<ScannerContext>();
		number= numberCounter++;
	}
	
	public ScannerContext(String target, PatternTraverser pattern, NodeListTraverser nodes) {
		this();
		clearBounds();
		
		this.nodes= nodes;
		this.pattern= pattern;
		this.target= target;
		fit=1.0;
		
		nextStrokePattern();
		
		if (ParameterSet.verbose)
			messages= new MessageStack();
	}
	
	ScannerContext minimalSpawn() {
		ScannerContext child= new ScannerContext();
		child.target= target;
		child.fit= fit;
		child.lastpoint= lastpoint;
		child.nodes= nodes.copy();
		if (messages!=null)
			child.messages= messages.copy();
		if (ParameterSet.tryVariations)
			offspring.add(child);
		return child;		
	}
	
	public ScannerContext spawn() {
		ScannerContext child= minimalSpawn();
		child.nodes= nodes.copy();
		child.bounds= bounds.copy();
		child.locators= locators.copy();
		if (hasMorePattern())
			child.pattern= new PatternTraverser(pattern);
		return child;
	}
	
	void clearBounds() {
		locators= new BiStack<BiStack<ScaledLocator>>();
		locators.addFirst(new BiStack<ScaledLocator>());
		bounds= new BiStack<BoundingBox>();		
	}
	
	public Collection<ScannerContext> getOffspring() {
		Collection<ScannerContext> result= offspring;
		if (!offspring.isEmpty())
			offspring= new LinkedList<ScannerContext>();
		return result;
	}
	
	public int searchedStrokeTypes() {
		if (pattern==null)
			return 0;
		else
			return ((StrokePattern) pattern.current()).getStrokeBits();
	}
	
	public boolean strokeIsConnected() {
		if (pattern==null)
			return false;
		else
			return ((StrokePattern)pattern.current()).isConnected();
	}
	
	public void nextStrokePattern() {
		nextStrokePattern(true);
	}
	
	public Collection<ScannerContext> nextStrokePattern(boolean doParasites) {
		Collection<ScannerContext> parasites= null;
		if (doParasites) 
			parasites= new LinkedList<ScannerContext>();
		while (pattern!=null && pattern.hasNext()) {
			ScannerContext context=nextPattern(doParasites);
			if (context!=null)
				parasites.add(context);
			if (pattern.current() instanceof StrokePattern)
				return parasites;
		}
		pattern= null;
		return parasites;
	}
	
	protected ScannerContext nextPattern(boolean doParasites) {
		ScannerContext variant=null;
		while (pattern.hasNext()) {
			if (pattern.isLastInGroup()) {
				pattern.next();
				endContext();
				// break if this context is stale
				if (getFit()<ParameterSet.minimumFit)
					break;
			} else {
				pattern.next();
				if (doParasites && bounds.size()<=2 && pattern.current() instanceof Parasite) {
					Rule p = ((Parasite) pattern.current()).getRule();
					variant= minimalSpawn();
					variant.clearBounds();
					variant.message("***** spawning from "+target);
					variant.target=p.getTarget();
					variant.pattern=new PatternTraverser(p.getPattern());
					while (! (variant.pattern.current() instanceof ReferencePattern) ||
							variant.pattern.current() instanceof MacroPattern) { 
						variant.nextPattern(false);
					}
					variant.expand(bounds.peek());					
					variant.pattern.skipChildren();
					variant.nextStrokePattern();
				}
				beginContext();
				return variant;
			}
		}
		return null;
	}

	protected void beginContext() {
		bounds.addFirst(new BoundingBox());
		if (pattern.current() instanceof ReferencePattern) {
			locators.addFirst(new BiStack<ScaledLocator>());
		} else if (pattern.current() instanceof ArgumentPattern) {
			locators.hide();
		}
	}
	
	protected void endContext() {
		BoundingBox box= bounds.poll();
		if (pattern.current() instanceof ReferencePattern) {
			locators.removeFirst();
		} else if (pattern.current() instanceof ArgumentPattern) {
			locators.unhide();
		}

		if (box.isActive()) {
			if (bounds.hasNext())
				expand(box);
			if (pattern.current().locators() != null) {
				locators.addFirst(locators.poll().copy());
				for (Locator l : pattern.current().locators()) {
					double lfit=1.0;
					if (l.getName()!= null) {
						ScaledLocator sl = new ScaledLocator(l, box);
						lfit = Math.min(lfit, getFit(sl));
						locators.peek().addFirst(sl);
					} else {
						if (bounds.size()<2)
							lfit = Math.min(lfit, ScaledLocator.shapeFit(l,box));
					}
					message("Locator ",l.getName()," fits "+lfit);
					applyFit(lfit);
				}
			}
		}
	}
	
	public void expand(Vector2d point) {
		bounds.addFirst(new BoundingBox(bounds.poll()));
		bounds.peek().expand(point);
		lastpoint= point;
	}

	public int compareTo(ScannerContext other) {
		if (getFit() < other.getFit())
			return 1;
		if (getFit() > other.getFit())
			return -1;
		if (number > other.number) 
			return 1;
		if (number < other.number)
			return -1;
		return 0;
	}

	public void expand(BoundingBox box) {
		bounds.addFirst(new BoundingBox(bounds.poll()));
		bounds.peek().expand(box);
	}

	public Vector2d getLastPoint() {
		return lastpoint;
	}
	
	protected double getFit(ScaledLocator loc) {
		double fit=1.0;
		for (ScaledLocator loc2 : locators.peek()) {
			if (loc.getName()==loc2.getName()) {
				fit=Math.min(fit, loc.getFit(loc2));
			}
		}
		return fit;
	}
	
	public void applyFit(double factor) {
		fit= fit*factor;
		if (factor !=1.0)
			message("New Fittness: "+fit);
	}
	
	public double getFit() {
		return fit;
	}
	
	public boolean hasMorePattern() {
		return pattern != null;
	}
	
	public boolean isComplete() {
		return !hasMorePattern() && nodes.isComplete();
	}
	
	public boolean isActive() {
		boolean a= nodes.isActive() && hasMorePattern();
		return a;
	}
	public void message(Object...objects) {
		if (messages!=null)
			messages.message(objects);
	}
	
	public String toString() {
		return "Context #"+number+"\n"+ messages.toString();
	}
}
