/* Tibi - Chinese handwriting recognition
 * Copyright (C) 2007 Stefan Dirnstorfer <stefan@dadim.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
package rules;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.Vector;

public class RuleSet implements Iterable<Rule> {
	Vector<Rule> rules;
	boolean resolved;
	
	public RuleSet() {
		rules= new Vector<Rule>();
	}
	
	public void add(Rule rule) {
		rules.add(rule);
		resolved=false;
	}

	public void resolveAll() {
		for (Rule r : rules) {
			try {
				if (r.getPattern() != null)
					resolveReferences(r.getPattern());
			} catch (Error e) {
				System.err.println("Error in line "+r.line+" :  "+e.getMessage());
			}
		}
		compress();
		resolved=true;
	}
	
	public void compress() {
		LinkedList<Rule> incremental= new LinkedList<Rule>();
		for (Rule r : rules) {
			if (r.isVisible()) {
				// search for the first referenced subpattern
				PatternTraverser tra= new PatternTraverser(r.getPattern());
				String pseudoRadical=null;
				while(tra.hasNext()) {
					Pattern cur= tra.next();
					if (cur instanceof StrokePattern) {
						break;
					}
					if (cur instanceof ReferencePattern && ! (cur instanceof MacroPattern)) {
						pseudoRadical= ((ReferencePattern)tra.current()).getName();
						break;
					}
				}
				// check if the first reference is visible, then infect
				if (pseudoRadical !=null) {
					Rule p= getRule(pseudoRadical);
					if (p.isVisible()) {
						p.pattern.add(new Parasite(r));
						incremental.add(r);
					}
				}
			}
		}
		for (Rule r : incremental) {
			r.visible=false;
		}
	}
	
	public void resolveReferences(Pattern pattern) {
		if (pattern instanceof ReferencePattern) {
			ReferencePattern refpat= (ReferencePattern) pattern;
			if (refpat.pattern == null) {
				Pattern pat= getPattern(refpat.name);
				if (pat== null)
					throw new Error("Unresolved reference for "+ refpat);
				refpat.pattern=pat;
			}
		}
		if (pattern instanceof Iterable) {
			for (Pattern p : ((Iterable<Pattern>) pattern)) {
				resolveReferences(p);
			}
		}
	}
		
	public Iterator<Rule> iterator() {
		if (!resolved)
			resolveAll();
		return rules.iterator();
	}

	public Rule getRule(String target) {
		for (Rule r: rules) {
			if (r.getTarget().equals(target))
				return r;
		}
		return null;		
	}
	
	public Pattern getPattern(String target) {
		Rule r= getRule(target);
		if (r==null)
			return null;
		else
			return r.getPattern();
	}
	
	public int getLineNumber(String target) {
		for (Rule r: rules) {
			if (r.getTarget().equals(target))
				return r.getLineNumber();
		}
		return -1;
	}
	
	public int visibleRuleCount() {
		int v=0;
		for (Rule r : rules) {
			if (r.isVisible()) v++;
		}
		return v;
	}
	
	public int totalRuleCount() {
		return rules.size();
	}
	
	public String toString() {
		return rules.toString();
	}
}
