/* Tibi - Chinese handwriting recognition
 * Copyright (C) 2007 Stefan Dirnstorfer <stefan@dadim.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
package util;

import java.util.Collection;
import java.util.Iterator;

/** Bidirectional stack. This stack behaves like a normal stack, but has the additional 
 * feature of hiding and unhiding stack entries. Hidden entries are put on a second invisible stack,
 * from where they can be retrieved and put on top of the visible stack with unhide.
 * <br/>
 * BiStack entries are stored in a single linked list with immutable item data. The creation of 
 * deep copies is therefore extremely efficient. 
 */
public class BiStack<T> implements Iterable<T>{

	Item up, down;
	int count;
	
	class Item {
		T entry;
		Item next;
		Item (T entry) {
			this.entry=entry;
		}
		Item (T entry, Item next) {
			this.entry=entry;
			this.next=next;
		}
	}
	
	/** Creates an empty BiStack.
	 */
	public BiStack() {
		count=0;
	}
	
	/** Returns a virtual deep copy, which can be modified without intereference of the original stack.
	 *  Entry data is not copied.
	 * 
	 * @return virtual deep copy.
	 */
	public BiStack<T> copy() {
		BiStack<T> c= new BiStack<T>();
		c.count= count;
		c.up= up;
		c.down= down;
		return c;
	}
	
	/** Deletes all entries from the current stack and
	 * replaces them with the entries from the collection in the original order.
	 * @param src Source of new entries.
	 */ 
	public void set(Iterable<T> src) {
		Item item= null;
		clear();
		for (T entry : src) {
			if (item== null) {
				item= new Item(entry);
				up= item;
			} else {
				item.next= new Item(entry);
				item= item.next;
			}
		count++;
		}
	}
	
	
	/** Deletes all entries from the stack.
	 *  Visible and invisible entries are deleted.
	 */
	public void clear() {
		count=0;
		up= null;
		down= null;
	}
	
	/** Returns an iterator of the stack entries.
	 * 
	 */
	public Iterator<T> iterator() {
		return new BiStackIterator<T>(this);
	}
	
	/** Adds an item on top of the stack.
	 * 
	 * @param item Item to push.
	 */	
	public void addFirst(T item) {
		up= new Item(item, up);
		count++;
	}
	
	/** Returns if there are more items on the visible stack.
	 * 
	 * @return 
	 */
	public boolean hasNext() {
		return up != null;
	}
	
	/**
	 * 
	 * @return first item on stack
	 */
	public T peek() {
		if (up != null) 
			return up.entry;
		else 
			return null;
	}
	
	public T poll() {
		T result= up.entry;
		removeFirst();
		return result;
	}
	
	public void removeFirst() {
		up= up.next;
		count--;
	}
	
	public void hide() {
		down= new Item(up.entry, down);
		removeFirst();
	}
	
	public void unhide() {
		addFirst(down.entry);
		down= down.next;
	}
	
	/** Returns the number of visible entries on the stack. 
	 * Hidden elemens are not counted.
	 * @return stack size
	 */
	public int size() {
		return count;
	}
}
